#include <SDL.h>
#include "plugin.h"
#include <vector>
#include <sys/stat.h>

int cprintf(char *fmt, ...);

#include "MusicPlayer.h"
#include "FileList.h"
#include "Fifo.h"


#ifdef _WIN32
#include <windows.h>
void *dlopen(const char *filename, int flag)
{
	return (void *)LoadLibraryA(filename);
}

void *dlsym(void *handle, const char *symbol)
{
	HINSTANCE h = (HINSTANCE)handle;
	return (void *)GetProcAddress(h, symbol);
}

#else
#include <dlfcn.h>

#define stricmp strcasecmp
#define strnicmp strncasecmp

#endif


//static int volume = 1000;

#ifdef _WIN32
#define BLOCK_SIZE 4096
#define BLOCK_COUNT 4
#else
#define BLOCK_SIZE 1024
#define BLOCK_COUNT 8
#endif

struct sound_plugin *current_plugin = NULL;


static short audiobuf[BLOCK_SIZE * BLOCK_COUNT];

void MusicPlayer::fill_audio(BlockFifo *af, int max)
{
//	fprintf(stderr, "%p\n", current_plugin);
	if(current_plugin && current_plugin->fill_buffer)
	{
		//SDL_LockAudio();

		int b = af->LeftToWrite();

//		fprintf(stderr, "Can fill %d blocks\n", b);

		if(b)
		{
			if(max && b > max) b = max;
			int rc = current_plugin->fill_buffer(audiobuf, BLOCK_SIZE * b);
			if(!rc)
			{
				if(current_plugin && current_plugin->close)
					current_plugin->close();
				current_plugin = NULL;
				fprintf(stderr, "DONE\n");
			}
			else
				af->Write(audiobuf, rc / BLOCK_SIZE);
		}
	
		//SDL_UnlockAudio();
	}
}

static char go_on = 1;
static char paused = 0;

int MusicPlayer::GetNextSamples(short *dest, int len)
{
	int blocks = (len+BLOCK_SIZE-1)/BLOCK_SIZE;

	if(blocks > audioFifo->LeftToRead())
		blocks = audioFifo->LeftToRead();

	audioFifo->Read(dest, blocks, true);
	return blocks * BLOCK_SIZE;
}

void MusicPlayer::audio_cb(void *userdata, Uint8 *stream, int len)
{
	if(paused)
	{
		memset(stream, 0, len);
		return;
	}

	MusicPlayer *mp = (MusicPlayer *)userdata;

	//short *samples = (short *)stream;
	//BlockFifo *af = (BlockFifo *)userdata;
	//fprintf(stderr, "%d bytes, %d blocks available\n", len, ps.audioFifo->LeftToRead());
	if(mp->audioFifo->LeftToRead())
	{
		//if(mp->heavy)
		//	mp->audioFifo->Read(stream, 1);
		//else
			mp->audioFifo->Read16(stream, 1, mp->vol * 0x10000 / 10);

		if(mp->blend)
		{
			int b = mp->blend;
			int l,r,i;
			short *s = (short *)stream;
			for(i=0; i<len/4; i++, s += 2)
			{
				r = (s[0] * b + s[1] * (256-b)) >> 8;
				l = (s[1] * b + s[0] * (256-b)) >> 8;
				s[0] = l;
				s[1] = r;
			}
		}


		mp->total_samples += len;
		//fprintf(stderr, "Playing\n");
	}
	else
		memset(stream, 0, len);

	if(!go_on)
		return;

	fill_audio(mp->audioFifo, 2);
}


int MusicPlayer::InitSound()
{
	SDL_AudioSpec desired, obtained;

	desired.freq = 22050;

	if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
		desired.format = AUDIO_S16MSB;
	else
		desired.format = AUDIO_S16LSB;

	desired.channels = 2;

	desired.samples = BLOCK_SIZE/4;

	audioFifo = new BlockFifo(BLOCK_SIZE, BLOCK_COUNT);

	desired.callback = audio_cb; 
	desired.userdata = this;

	fprintf(stderr, "Open Audio\n");


	if(SDL_OpenAudio(&desired, &obtained) < 0)
	{
      	cprintf("Couldn't open SDL audio: %s\n", SDL_GetError());
		return 0;
	}


	SDL_PauseAudio(0);

	return 1;
}

#ifdef STATIC

extern "C" {
sound_plugin *sidplay_init_sound_plugin();
sound_plugin *modplay_init_sound_plugin();
sound_plugin *uade_init_sound_plugin();
}

#endif

static int is_ext(const char *s, const char *ext)
{
	const char *sext = strrchr(s, '.');
	if(!sext)
		sext = "";
	return (stricmp(ext, sext) == 0);
}


int MusicPlayer::CanHandle(const char *name)
{
	for(unsigned int i=0; i<sound_plugins.size(); i++)
		if(sound_plugins[i]->can_handle(name))
		{
			return 1+1;
		}
	return 0;
}


MusicPlayer::MusicPlayer()
{
#ifdef STATIC

	struct sound_plugin *plugin;

	plugin = sidplay_init_sound_plugin();
	sound_plugins.push_back(plugin);
	plugin = modplay_init_sound_plugin();
	sound_plugins.push_back(plugin);

#ifndef _MSC_VER
	plugin = uade_init_sound_plugin();
	sound_plugins.push_back(plugin);
#endif

#else

	FileList *flist = new FileList("plugins");

	cprintf("%p %d\n", flist, flist ? flist->Size() : -1);

	for(int i=0; i<flist->Size(); i++)
	{
		const char *fn = flist->GetFileName(i);
		void *ptr = dlopen(fn, 0);

		cprintf("Plugin %s:", fn);

		if(ptr)
		{
			struct sound_plugin *(*init)() = (struct sound_plugin *(*)())dlsym(ptr, "init_sound_plugin");

			if(init)
			{
				struct sound_plugin *plugin = init();
				sound_plugins.push_back(plugin);

				cprintf(" Loaded\n");
			}
			else
				cprintf(" Init failed!\n");
		}
		else
			cprintf(" Not valid sound plugin\n");
	}
#endif

	current_plugin = NULL;
	vol = 10;
	subtune = 0;
	total_samples = 0;
	blend = 0;
	heavy = 0;

	InitSound();

}

//const char *basename(const char *fname);

#ifdef _WIN32
const char *basename(const char *fname)
{
	const char *ptr = &fname[strlen(fname)-1];

	while(ptr > fname && *ptr != ':' && *ptr != '\\' && *ptr != '/')
		ptr--;

	if(ptr != fname)
		ptr++;

	return ptr;
}
#endif

char *module = NULL;

bool MusicPlayer::PlaySong(const char *fname)
{
	for(unsigned int i=0; i<sound_plugins.size(); i++)
	{
		if(sound_plugins[i]->can_handle(fname) && sound_plugins[i]->init_file)
		{
			int rc = -1;

			cprintf("plugin\n");
			go_on = 0;
			SDL_LockAudio();
			if(current_plugin && current_plugin->close)
				current_plugin->close();

			if(module)
				free(module);
			module = NULL;

			current_plugin = sound_plugins[i];
			if(current_plugin->init_file)
				rc = current_plugin->init_file((char *)fname);
			else
			if(current_plugin->init_data)
			{
				struct stat ss;
				memset(&ss, 0, sizeof(struct stat));
				if(stat(fname, &ss) == 0)
				{
					char *tmp = (char *)malloc(ss.st_size);
					FILE *fp = fopen(fname, "rb");
					if(fp)
					{
						fread(tmp, 1, ss.st_size, fp);
						fclose(fp);
						rc = current_plugin->init_data((char *)fname, tmp, ss.st_size);
						free(tmp);
					}
				}
			}
			else
				rc = -1;

			fprintf(stderr, "Play Song = %d\n", rc);

			if(rc == 0)
			{
				total_samples = 0;
				subtune = current_plugin->tune;
				audioFifo->Reset();
				fill_audio(audioFifo, 0);
				if(!current_plugin)
					rc = -1;
				else
				{
					heavy = (strcmp(current_plugin->plugname, "uade") == 0);
					if(!strlen(current_plugin->name))
					{
						strcpy(temp_name, basename(fname));
						current_plugin->name = temp_name;
					}
				}
			}
			go_on = 1;
			SDL_UnlockAudio();
			return (rc == 0);
		}
	}
	return false;
}

bool MusicPlayer::PlaySong(const char *name, char *data, int size)
{
	for(unsigned int i=0; i<sound_plugins.size(); i++)
	{
		if(sound_plugins[i]->can_handle(name) && sound_plugins[i]->init_data)
		{
			int rc = -1;

			cprintf("plugin\n");
			go_on = 0;
			SDL_LockAudio();
			if(current_plugin && current_plugin->close)
				current_plugin->close();

			module = (char *)malloc(size);
			memcpy(module, data, size);

			current_plugin = sound_plugins[i];
			rc = current_plugin->init_data((char *)name, module, size);

			fprintf(stderr, "Play Song from data= %d\n", rc);

			if(rc >= 0)
			{
				total_samples = 0;
				subtune = current_plugin->tune;
				audioFifo->Reset();
				fill_audio(audioFifo, 0);
				if(!current_plugin)
					rc = -1;
				else
				{
					heavy = (strcmp(current_plugin->plugname, "uade") == 0);
					if(!strlen(current_plugin->name))
					{
						strcpy(temp_name, basename(name));
						current_plugin->name = temp_name;
					}
				}

			}

			go_on = 1;
			SDL_UnlockAudio();
			return (rc >= 0);
		}
	}
	return false;
}

void MusicPlayer::SetVolume(int v)
{ 
	if(v < 0)
		v = 0;
	if(v > 10)
		v = 10;
	vol = v;
}


void MusicPlayer::SetTune(int n)
{
	if(!current_plugin)
		return;
	if(!current_plugin->set_position)
		return;

	if(n < 0)
		n = 0;
	if(n >= current_plugin->subtunes)
		n = current_plugin->subtunes-1;


	if(subtune != n)
	{
		//cprintf("tune %d\n", n);
		total_samples = 0;
		subtune = n;
		go_on = 0;
		SDL_LockAudio();
		current_plugin->set_position(0, n);
		audioFifo->Reset();
		fill_audio(audioFifo, 0);
		go_on = 1;
		SDL_UnlockAudio();
	}
}



int MusicPlayer::Pause()
{
	if(paused)
		paused = 0;
	else
		paused = 1;
	return paused;
}
