#ifndef _SCREEN_H_
#define _SCREEN_H_

struct SDL_Surface;

#include <pogo/font.h>

#include <stdarg.h>

#define TO_RGB16(r,g,b) ( ((r << 8) & 0xF800) | ((g << 3) & 0x07E0) | (b >> 3) )
#define UINT2RGB16(x) (((x>>19)&0x1F)<<11) | (((x>>10)&0x3F)<<5) | ((x>>3)&0x1F)

class IFileList;

struct ScreenSettings
{
	int width;
	int height;
	char *font0;
	char *font1;
	uint32 bgcol;
	uint32 textcol;
	uint32 beveldark;
	uint32 bevelmid;
	uint32 bevellight;
};

class Screen
{
public:
	Screen();
	void Init();

	void CreateConsole(int x, int y, int w, int h);

	int PrintV(char *fmt, va_list vl);
	int Print(char *fmt, ...);

	void *Lock() { SDL_LockSurface(screen); return screen->pixels; }
	void Unlock() { SDL_UnlockSurface(screen); }
	void Update() { SDL_UpdateRect(screen, 0, 0, 0, 0); }

	int GetFontHeight() { return font1->height; }

	void Resize(int w, int h);

	ScreenSettings *GetSettings();
	void SetSettings(ScreenSettings *settings);

protected:

	ScreenSettings settings;

	//Font *fonts[4];
	Font *font0;
	Font *font1;
	int con_fd;
	int mainwin;
	SDL_Surface *screen;


	friend class Pane;
};


class Pane
{
public:
	Pane(int x, int y, int w, int h, Screen *scr = NULL);
	virtual int Text(const char *t, int x, int y);
	virtual void Render();
	virtual void Move(int x, int y);
	virtual void Resize(int w, int h);
	virtual void Clear(bool bevel);

	virtual bool Update();

	virtual int Width() { return m_w; }
	virtual int Height() { return m_h; }

	enum { REDRAW = 1, RESIZE };

protected:
	Font *font;
	SDL_Surface **screen;
	int dirty;
	int  m_x;
	int  m_y;
	int  m_w;
	int  m_h;

	Font *fonts[4];
	int con_fd;

};


class TextDisplay : public Pane
{
public:
	TextDisplay(int x, int y, int w, int h, Screen *scr = NULL);
	void PutText(int x, int y, const char *str);

	//bool Update();
	void Clear(bool bevel);

protected:
	int conwin;
};

class ListPane : public Pane
{
public:
	ListPane(int x, int y, int w, int h, Screen *scr = NULL);

	//int Text(const char *t, int x, int y);

	void SetList(IFileList *fl)
	{ 
		dirty = true;
		file_list = fl;
		marked = 0;
		start = 0;
	}

	IFileList *GetList() { return file_list; }

	bool Update();
	void DoCommand(int cmd);

	int GetMarked() { return marked; }
	void SetMarked(int m) { marked = m; }

	void Resize(int w, int h);

	enum {NOTHING, UP, DOWN, PGUP, PGDN};

protected:

	int con_fd;
	int conwin;
	IFileList *file_list;
	int marked;
	int oldmarked;
	int start;
	int lines;
	int last_cmd;
	int delay;
};

class LedControl : public Pane
{
public:
	LedControl(char *fname, int x, int y, unsigned int dcol = 0x003000, unsigned int lcol = 0x80FF80, Screen *scr = NULL);

	void SetLed(int led, int val);
	int GetLed(int led);

	bool Update();

	enum {
		BLACK, SPEAKER, VOLUME,
		UNUSED = 12,
		PLAYING = 13, SONG, SOLID, REPEAT, SHUFFLE, 
		NUMBERS = 18, LENGTH = 18,
		SONGNO = NUMBERS + 15*5,
		MAXSONG = NUMBERS + 15*7,
		BAR0 = 153,
		BAR1 = 165,
		EQ0 = 177,
		EQ1 = 182,
		EQ2 = 187,
		EQ3 = 192,
		BUSY = 197, // 6 long
	};
protected:

	void SetNumbers(int led, int val, int digs);

	unsigned short clut[256];
	unsigned short dark;
	unsigned short light;


	SDL_Surface *display;

};

class SongInfoPane : public Pane
{
public:
	SongInfoPane(int x, int y, int w, int h, Screen *scr = NULL);
	bool Update();

	void SetArtist(const char *a);
	void SetCopyright(const char *c);
	void SetName(const char *n);
	void SetFormat(const char *f);
	void SetLength(int l);

protected:

	char name[64];
	char author[64];
	char copyright[64];
	char format[64];
	int length;

};

int cprintf(char *fmt, ...);


extern Screen g_screen;

#endif
