/*  Copyright 2006 Jonas Minnberg

    This file is part of OldPlay - a portable, multiformat musicplayer.

    OldPlay is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    OldPlay is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OldPlay; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef _SCREEN_H_
#define _SCREEN_H_

struct SDL_Surface;

#include <pogo/font.h>
#include <string.h>
#include <stdarg.h>

#define TO_RGB16(r,g,b) ( ((r << 8) & 0xF800) | ((g << 3) & 0x07E0) | (b >> 3) )
#define UINT2RGB16(x) (((x>>19)&0x1F)<<11) | (((x>>10)&0x3F)<<5) | ((x>>3)&0x1F)

class IFileList;

class Screen
{
public:

	enum {COL_BG, COL_TEXT, COL_BEVEL_DARK, COL_BEVEL_MID, COL_BEVEL_LIGHT,
		COL_SONG0, COL_SONG1, COL_MARK, COL_FLAG, COL_DIR,
		COL_LCD_BACK, COL_LCD_DARK, COL_LCD_LIGHT,
		COL_EQ_LOW, COL_EQ_MID, COL_EQ_HIGH, COL_EQ_BG,
		COL_BROWSER_TITLE0, COL_BROWSER_TITLE1, COL_PLAYLIST_TITLE0, COL_PLAYLIST_TITLE1,
		MAX_COL};

	static const char *ColorNames[];

	struct DisplaySettings
	{
		int width;
		int height;
		char *font0;
		char *font1;
		char *background;

		uint32 colors[MAX_COL];
/*
		uint32 bgcol;
		uint32 textcol;
		uint32 beveldark;
		uint32 bevelmid;
		uint32 bevellight;
		uint32 songcol0;
		uint32 songcol1;
		uint32 markcol;
		uint32 flagcol;
		uint32 dircol;
*/
	};

	Screen();
	void Init(DisplaySettings *s);
	int Width() { return settings.width; }
	int Height() { return settings.height; }
	
	void ScreenShot(const char *fname);

	void CreateConsole(int x, int y, int w, int h);

	int PrintV(char *fmt, va_list vl);
	int Print(char *fmt, ...);

	void *Lock() { SDL_LockSurface(screen); return screen->pixels; }
	void Unlock() { SDL_UnlockSurface(screen); }
	void Update() { SDL_UpdateRect(screen, 0, 0, 0, 0); }


	//int GetFontHeight() { return font1->height; }

	Font *GetPropFont() { return font1; }

	void Resize(int w, int h);

	//ScreenSettings *GetSettings();
	//void SetSettings(ScreenSettings *settings);

protected:

	void Setup();

	DisplaySettings settings;

	//Font *fonts[4];
	Font *font0;
	Font *font1;
	int con_fd;
	int mainwin;
	SDL_Surface *screen;


	friend class Pane;
};


class Pane
{
public:
	Pane(int x, int y, int w, int h, Screen *scr = NULL);
	virtual int Text(const char *t, int x, int y);
	virtual void Render();
	virtual void Move(int x, int y);
	virtual void Resize(int w, int h);
	virtual void Clear(bool bevel);

	virtual bool Update();

	virtual int Width() { return m_w; }
	virtual int Height() { return m_h; }

	enum { REDRAW = 1, RESIZE };

protected:
	Font *font;
	SDL_Surface **screen;
	int dirty;
	int  m_x;
	int  m_y;
	int  m_w;
	int  m_h;

	Screen::DisplaySettings settings;

	Font *fonts[4];
	int con_fd;

};


class TextDisplay : public Pane
{
public:
	TextDisplay(int x, int y, int w, int h, Screen *scr = NULL);
	void PutText(int x, int y, const char *str);

	//bool Update();
	void Clear(bool bevel);

protected:
	int conwin;
};

class ListPane : public Pane
{
public:
	ListPane(int x, int y, int w, int h, Screen *scr = NULL);

	//int Text(const char *t, int x, int y);

	void SetList(IFileList *fl)
	{ 
		dirty |= REDRAW;
		file_list = fl;
//		marked = 0;
		//start = 0;
	}

	IFileList *GetList() { return file_list; }

	void Render();
	bool Update();
	void DoCommand(int cmd);

	//int GetMarked() { return marked; }
	//void SetMarked(int m) { marked = m; }

	void Resize(int w, int h);

	void SetTitle(char *t) { strcpy(title, t); }
	void SetTitleColor(int col0, int col1)
	{ 
		titlecol0 = col0; titlecol1 = col1;
	}

	enum {NOTHING, UP, DOWN, PGUP, PGDN};

protected:

	char title[128];
	int titlecol0;
	int titlecol1;

	int con_fd;
	int conwin;
	IFileList *file_list;
	//int marked;
	//int oldmarked;
	//int start;
	int lines;
	int last_cmd;
	int delay;
	int last_change;
};

class LedControl : public Pane
{
public:
	LedControl(char *fname, int x, int y, Screen *scr = NULL);

	void SetLed(int led, int val);
	int GetLed(int led);

	bool Update();

	enum {
		BLACK, SPEAKER, VOLUME,
		HOLD = 12,
		PLAYING = 13, SONG, SOLID, REPEAT, SHUFFLE, 
		NUMBERS = 18, LENGTH = 18,
		SONGNO = NUMBERS + 15*5,
		MAXSONG = NUMBERS + 15*7,
		BAR0 = 153,
		BAR1 = 165,
		EQ0 = 177,
		EQ1 = 182,
		EQ2 = 187,
		EQ3 = 192,
		BUSY = 197, // 6 long
		VOLT = 203,
	};
protected:

	void SetNumbers(int led, int val, int digs);

	unsigned short clut[256];
	unsigned short dark;
	unsigned short light;


	SDL_Surface *display;

};

class SongInfoPane : public Pane
{
public:
	SongInfoPane(int x, int y, int w, int h, Screen *scr = NULL);
	bool Update();

	void SetArtist(const char *a);
	void SetCopyright(const char *c);
	void SetName(const char *n);
	void SetFormat(const char *f);
	void SetLength(int l);

protected:

	char name[64];
	char author[64];
	char copyright[64];
	char format[64];
	int length;
};

int cprintf(char *fmt, ...);


extern Screen g_screen;

#endif
